//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  expr.cpp
//! \brief Parse expressions using CRef, CContext, CObj, virtual CNode

#include "defs.hpp"

#include <boost/format.hpp>
using boost::format;

#include "../trivia/vector_ops.hpp"

#include "olf.hpp"
#include "mem.hpp"
#include "obj.hpp"
#include "expr.hpp"
#include "xax_lex.hpp"


//**************************************************************************************************
//*  CContext
//**************************************************************************************************


//! Constructor. Sets k,j,i to arguments. Sets other members to defaults.

CContext::CContext( int _k, int _j, int _i )
    : k(_k), j(_j), i(_i),                     // set from arguments
      vt(0), dim(_1), want_error(false), nv(0) // can be changed later
{
    if ( k==(int)-1 )
        return; // not sure whether k=-1 makes sense
    const COlo *f = SMem::instance()->mem_get(k);
    if ( j==(int)-1 ) {
        if ( f->nJ()==1 )
            j = 0;
    } else if ( j>=f->nJ() )
        throw "invalid evaluation context, j="+S(j)+">=nJ="+S(f->nJ());

    const COld *fd = dynamic_cast<const COld*>(f);
    if ( !fd ) {
        if ( i!=(int)-1 )
            throw S("invalid evaluation context, i set for curve file");
        return;
    }

    if ( j==(int)-1 )
        return;
    int ni = fd->nPts(j);
    if ( i==(int)-1 ) {
        if ( ni==1 )
            i = 0;
    } else if ( i>=ni )
        throw "invalid evaluation context for j="+S(j)+": i="+
            S(i)+">=ni="+S(ni);
}


//! Sets request of a vector of doubles, corresponding to input points 't'.

void CContext::request_VT( const vector<double> *_vt )
{
    dim = _VT;
    nv = _vt->size();
    vt = _vt;
    if( !triv::is_equidist( &vt_step, *vt ) || vt_step<0 )
        vt_step = 0;
}

//! Sets request of a vector of _nv indices.

void CContext::request_VI( int _nv )
{
    dim = _VI;
    nv = _nv;
}

//! Returns textual representation of the class instance.

string CContext::context_info() const
{
    string out = str( format( "[%i,%i," ) % k % j );
    if ( dim==_1 )
        out += str( format( "%i]" ) % i );
    else if ( dim==_VI )
        out += str( format( "{%i}]" ) % nv );
    else if ( dim==_VT )
        out += str( format( "%i*t]" ) % nv );
    if ( want_error )
        out += "+-";
    return out;
}


//**************************************************************************************************
//*  CRef (references like y[k,j,0] or p3[2] or ...)
//**************************************************************************************************


//! Returns file index k in given context.

int CRef1::get_k( const CContext& ctx ) const
{
    int k;
    if ( tk ) {
        RObj tmp = tk->tree_val( ctx );
        RObjInt pi = PCAST<const CObjInt>(tmp);
        if ( !pi )
            throw "index K=" + tk->tree_info() + "->" + tmp->to_s() + " is not integer";
        k = pi->val;
    } else {
        k = ctx.k;
    }
    if ( k==(int)-1 )
        throw S("missing or invalid k");
    SMem::instance()->mem_get( k ); // to assert existence of file
    return k;
}


//! Returns true if index k has an explicit dependence on i.

bool CRef1::k_depends_on_i() const
{
    return tk && tk->i_dependent();
}


//! Returns string representation of the class instance.

string CRef1::ref_info() const
{
    return "[" + ( tk ? tk->tree_info() : "K" ) + "]";
}


//! Returns spectrum index j in given context.

int CRef2::get_j( const CContext& ctx, int nj ) const
{
    int j;
    if ( tj ) {
        RObj tmp = tj->tree_val( ctx );
        RObjInt pi = PCAST<const CObjInt>(tmp);
        if ( !pi )
            throw "index J=" + tj->tree_info() + "->" + tmp->to_s() + " is not integer";
        j = pi->val;
    } else {
        j = ctx.j;
    }
    if ( j==-1 )
        throw S("index J=-1 (not defined?)");
    else if ( j<0 || j>=nj )
        throw "index J=" + S(j) + " is out of bounds";
    return j;
}


//! Returns true if index j has an explicit dependence on i.

bool CRef2::j_depends_on_i() const
{
    return tj && tj->i_dependent();
}


//! Returns string representation of the class instance.

string CRef2::ref_info() const
{
    return "[" +
        ( tk ? tk->tree_info() : "K" ) + "," +
        ( tj ? tj->tree_info() : "J" ) + "]";
}


//! Returns string representation of the class instance.

string CRef3::ref_info() const
{
    return "[" +
        ( tk ? tk->tree_info() : "K" ) + "," +
        ( tj ? tj->tree_info() : "J" ) + "," +
        ( ti ? ti->tree_info() : "I" ) + "]";
}


//**************************************************************************************************
//*  CNode: low-level tree evaluation
//**************************************************************************************************


//! Evaluates the node, and returns a non-negative integer.

int CNode::tree_val_idx( const CContext& ctx, const string& name ) const
{
    RObj val = tree_val( ctx );
    RObjInt pi = PCAST<const CObjInt>(val);
    if ( !pi )
        throw "index "+name+"="+val->to_s()+" is not integer";
    int ret = pi->val;
    if( ret<0 )
        throw "index "+name+"="+val->to_s()+" is negativ";
    return ret;
}


//! Evaluates the node, and returns a list of non-negative integers.

RObjVecInt CNode::to_index_list( int siz ) const
{
    CContext ctx;
    int nret = siz;
    ctx.request_VI( nret );
    ctx.want_error = false;
    RObj val = tree_val( ctx );
    if ( RObjInt ri = PCAST<const CObjInt>( val ) ) {
        PObjVecInt ret( new CObjVecInt );
        int ival = ri->to_i();
        if( ival<0 || ival>=siz )
            return nullptr;
        ret->v.push_back( ival );
        return ret;
    } else if ( RObjVecInt riv = PCAST<const CObjVecInt>( val ) ) {
        for( int ival: riv->v )
            if( ival<0 || ival>=siz )
                return nullptr;
        return riv;
    } else {
        return nullptr;
    }
}

//**************************************************************************************************
//*  CNode: public interface to tree evaluation
//**************************************************************************************************

//! Evaluates the node in a scalar context.

RObj CNode::tree_point_val( int k, int j, int i ) const
{
    CContext ctx( k, j, i );
    ctx.want_error = true;
    return tree_val( ctx );
}


//! Evaluates the node, and returns a double. OBSOLET!?

double CNode::tree_point_dbl( int k, int j, int i ) const
{
    CContext ctx( k, j, i );
    ctx.want_error = false;
    RObj val = tree_val( ctx );
    RObjNum pr = PCAST<const CObjNum>(val);
    if ( !pr )
        throw S("Tree point value has unexpected type");
    return pr->to_r();
}


//! Evaluates the node, and sets vectors *ret and *dret according to the result.

void CNode::tree_vec_val( vector<double> *ret, vector<double> *dret, int k, int j ) const
{
    CContext ctx( k, j );
    int nret = ret->size();
    ctx.request_VI( nret );
    ctx.want_error = (bool)dret;
    RObj val = tree_val( ctx );
    if        ( RObjNum pr = PCAST<const CObjNum>(val) ) {
        ret->assign( nret, pr->to_r() );
        if( dret && pr->has_err() )
            dret->assign( nret, pr->to_dr() );

    } else if ( RObjVecInt pv = PCAST<const CObjVecInt>(val) ) {
        for( int i=0; i<nret; ++i )
            (*ret)[i] = pv->to_r(i);
    } else if ( RObjVecDbl pv = PCAST<const CObjVecDbl>(val) ) {
        *ret = pv->v;
        if( dret ) {
            if ( RObjVecEnu pdv = PCAST<const CObjVecEnu>(val) )
                *dret = pdv->dv;
            else
                dret->clear();
        }
    } else
        throw S("BUG in tree_vec_val: unexpected return type");
}

//! Returns a node obtained by parsing the given expression.

RObj CNode::eval( string expr )
{
    RNode T = user_xaxparse( expr.c_str() );
    return T->tree_point_val();
}
