//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  fbase.cpp
//! \brief Collection of basic and arithmetic functions.


#include <cfloat>
#include <gsl/gsl_sf.h>

#include "../trivia/rng.hpp"
#include <kww.h>

#include "defs.hpp"
#include "func.hpp"
#include "fregistry.hpp"

#include <cerf.h> // include after boost::shared_ptr, otherwise use of 'I' in template clashes

const double PI    = M_PI;    //4*atan(1.);
const double TWOPI = 2*M_PI; //8*atan(1.);

static triv::RandomNumberGenerator *rng = nullptr;


//**************************************************************************************************
//*  Functions of one argument
//**************************************************************************************************

template<class A>
void func_pop(A v) {;}

double func_to_d(int v) { return v; }
double func_to_d(double v) { return v; }
double func_to_d(string s)
{
    double ret;
    if ( triv::any2dbl( s, &ret ) )
        return ret;
    throw "Cannot convert "+s+" to floating-point number";
}

template<class A>
int func_not(A v) { return !v; }

template<class A>
A func_neg(A v) { return -v; }
void func_neg(double& r, double& dr, double v, double dv) { r=-v; dr=dv; }

double func_ln(double v) { return v>0 ? log(v) : 0; }
void func_ln(double& r, double& dr, double v, double dv)
{ if(v>0) { r=log(v); dr=dv/v; } else { r=0; dr=0; } }

double func_lg(double v) { return v>0 ? log10(v) : 0; }
void func_lg(double& r, double& dr, double v, double dv)
{ if(v>0) { r=log10(v); dr=dv/v/log(10); } else { r=0; dr=0; } }

double func_sqrt(double v) { return v>=0 ? sqrt(v) : 0; }
void func_sqrt(double& r, double& dr, double v, double dv)
{ if(v>0) { r=sqrt(v); dr=dv/2/r; } else { r=0; dr=0; } }

double func_abs(double v) { return fabs(v); }
int    func_abs(int v)    { return abs(v); }
void   func_abs(double& r,double& dr, double v,double dv) { r=fabs(v); dr=dv; }

template<class A>
A func_sign(A v) {
    if      ( v<0 )
        return -1;
    else if ( v>0 )
        return 1;
    else
        return 0;
}
void func_sign(double& r,double& dr, double v,double dv) { r=func_sign(v); dr=dv; }

double func_exp(double v) { return exp(v); }
void func_exp(double& r, double& dr, double v, double dv) { r=exp(v); dr=dv*r; }

double func_sin(double v) { return sin(v); }
void func_sin(double& r, double& dr, double v, double dv) { r=sin(v); dr=dv*fabs(cos(v)); }
double func_cos(double v) { return cos(v); }
void func_cos(double& r, double& dr, double v, double dv) { r=cos(v); dr=dv*fabs(sin(v)); }
double func_tan(double v) { return tan(v); }
double func_cot(double v) { double t=tan(v); return t==0?0:1/t;}

double func_sind(double v) { return sin(v*TWOPI/360.); }
double func_cosd(double v) { return cos(v*TWOPI/360.); }
double func_tand(double v) { return tan(v*TWOPI/360.); }
double func_cotd(double v) { double t=tan(v*TWOPI/360.);  return t==0?0:1/t; }

double func_asin(double v) { return (-1<=v && v<=1) ? asin(v) : 0; }
double func_acos(double v) { return (-1<=v && v<=1) ? acos(v) : 0; }
double func_atan(double v) { return atan(v); }
double func_acot(double v) { return v!=0 ? atan(1/v) : TWOPI/4; }

double func_asind(double v) { return (-1<=v && v<=1) ? 360./TWOPI*asin(v) : 0; }
double func_acosd(double v) { return (-1<=v && v<=1) ? 360./TWOPI*acos(v) : 0; }
double func_atand(double v) { return 360./TWOPI*atan(v); }
double func_acotd(double v) { return v!=0 ? 360./TWOPI*atan(1/v) : 90; }

double func_sinh(double v) { return sinh(v); }
double func_cosh(double v) { return cosh(v); }
double func_tanh(double v) { return tanh(v); }
double func_coth(double v) { double t=tanh(v); return t==0?0:1/t;}

double func_j0(double v) { return v==0 ? 1 : sin(v)/v; }
double func_j1(double v) { return v==0 ? 1 : (sin(v)/v-cos(v))/v; }

double func_fac(double v) { return v>-.5 ? tgamma(round(v+1)) : 0; }
double func_cata(double v) { // Catalan number ?
    if ( v<=-.5 )
        return 0;
    double x = round(v);
    return exp( lgamma( 2*x+1 ) - lgamma( x+1 ) -
                lgamma( x+2 ) ); }

double func_gamma(double v) { return v>0 ? tgamma(v) : 0; }
double func_lgamma(double v) { return v>0 ? lgamma(v) : 0; }
// could be extended to non-integer negative arguments...

double func_debye1(double v) { return v>=0 ? gsl_sf_debye_1(v) : 0; }
double func_debye3(double v) { return v>=0 ? gsl_sf_debye_3(v) : 0; }
double func_debye_msd(double v) { // mean squared displacement
                                  // without prefactors
                                  // without zero temperature oscillations
    return v>1e-2 ? 2*v*gsl_sf_debye_1(1/v) : 0; }
double func_debye_ui(double v) { // internal energy
    return v>=0 ? 3*v*gsl_sf_debye_3(1/v) : 0; }
double func_debye_cv(double v) { // heat capacity
    double eps=1e-12;
    return v>=eps ? (func_debye_ui(v+eps)-func_debye_ui(v-eps))/(2*eps) : 0; }

double func_erfP(double v) { return 1-gsl_sf_erf_Q(v); }
double func_erfQ(double v) { return gsl_sf_erf_Q(v); }
double func_erf(double v) { return gsl_sf_erf(v); }
double func_erfc(double v) { return gsl_sf_erfc(v); }
double func_erfi(double v) { return erfi(v); }

double func_dawson(double v) { return gsl_sf_dawson(v); }

double func_sinc(double v) { return v==0 ? 1 : sin(v)/v; }

int func_ceil(double v)  { return (int)(ceil(v)+0.5); }
int func_floor(double v) { return (int)(floor(v)+0.5); }
int func_nint(double v)  { return (int)(round(v)+0.5); }

// Ad hoc for uses in physics:

double func_lndiehl( double a ) {
    // ln( \int dw gauss(w,gausswid) cauchy(w,cauchywid) / gauss(0,gausswid) )
    // argumnt a = cauchywid/gausswid
    return SQR(a)/2 + gsl_sf_log_erfc( a/sqrt(2) );
}

double func_diehl( double a ) {
    // \int dw gauss(w,gausswid) cauchy(w,cauchywid) / gauss(0,gausswid)
    // argumnt a = cauchywid/gausswid
    return exp( func_lndiehl(a) );
}

int func_exit( int a ) { exit(a); return 0; }

int func_exit_if( int a, string s ) { if(a) { cout << s << "\n"; exit(1); } return 0; }
int func_exit_unless( int a, string s ) { return func_exit_if( !a, s ); }

//**************************************************************************************************
//*  Functions of two arguments
//**************************************************************************************************

//double func_pm(double v, double a) { return v; }
void func_pm(double& r, double& dr, double v, double a){ r=v; dr=a; }

template<class A>
A func_add(A v, A a) { return v+a; }
void func_add(double& r, double& dr, double v, double dv, double a, double da)
{ r=v+a; dr=sqrt(dv*dv+da*da); }

template<class A>
A func_sub(A v, A a) { return v-a; }
void func_sub(double& r, double& dr, double v, double dv, double a, double da)
{ r=v-a; dr=sqrt(dv*dv+da*da); }

template<class A>
A func_mul(A v, A a) { return v*a; }
void func_mul(double& r, double& dr, double v, double dv, double a, double da)
{ r=v*a; dr=sqrt(SQR(v*da)+SQR(a*dv)); }

double func_div(double v, double a) { return a!=0 ? v/a : 0; }
void func_div(double& r, double& dr, double v, double dv, double a, double da)
{ if(a==0) { r=0; dr=0; } else { r=v/a; dr=sqrt(SQR(dv/a)+SQR(v*da/a/a)); } }

int func_idiv(int v, int a) { return a!=0 ? v/a : 0; }
int func_idiv(double v, int a) { return a!=0 ? (int)(ceil(v)+0.5)/a : 0; }

int func_mod(int v, int a) { return v%a; }
double func_mod(double v, double a) { return fmod(v,a); }

double func_pow(double a, double b)
{ return ( a>0 || (a==0&&b>0) || b==(int)b ) ? pow(a,b) : 0; }
void func_pow(double& r, double& dr, double a, double da, double b, double db)
{
    if ( a>0 ) {
        r = pow(a,b);
        dr = r*b/a*sqrt(SQR(da)+SQR(db));
    } else if ( a==0 && b>0 ) {
        r = pow(a,b);
        dr = b*pow(a,b-1)*db;
    } else if ( b==(int)b ) {
        r = pow(a,b);
        dr = r*b/a*da;
    } else {
        r = 0;
        dr = 0;
    }
}

template<class A>
int func_eq (A v, A a) { return v == a; }
template<class A>
int func_ne (A v, A a) { return v != a; }
template<class A>
int func_gt (A v, A a) { return v > a; }
template<class A>
int func_ge (A v, A a) { return v >= a; }
template<class A>
int func_lt (A v, A a) { return v < a; }
template<class A>
int func_le (A v, A a) { return v <= a; }
template<class A>
int func_and(A v, A a) { return v && a; }
template<class A>
int func_or (A v, A a) { return v || a; }
template<class A>
int func_xor(A v, A a) { return (bool) v ^ (bool) a; }

double func_over(double v, double a) {
    return  ( 0<=v && 0<=a && a<=v && v <= 400 ) ?
        tgamma(v+1)/tgamma(a+1)/tgamma(v-a+1) :
        0; }

template<class A>
A func_min (A v, A a) { return v<a ? v : a; }
template<class A>
A func_max (A v, A a) { return v>a ? v : a; }

double func_gauss (double x, double s) {
    return s==0 ? 0 : exp(-x*x/(2*s*s))/sqrt(TWOPI*s*s); }
double func_gaussnn (double x, double s) { // Gauss not normalized
    return s==0 ? 0 : exp(-x*x/(2*s*s)); }
double func_cauchy (double x, double s) {
    return s==0 ? 0 : fabs(s)/PI/(x*x+s*s); }

double func_ran (double v, double a) {
    // a random number between v and a :
    if (!rng)
        rng = new triv::RandomNumberGenerator();
    return rng->uniform( func_min(v,a), func_max(v,a) ); }

// Havriliak-Negami: unchecked formulae from disreputable source:

double aux_phi_havneg (double v, double a){
    return atan( pow(v,a)*sin(M_PI*a/2)/(1+pow(v,a)*cos(M_PI*a/2)) ); }

double func_re_havneg (double v, double a, double g) {
    return pow( 1 + 2*pow(v,a)*cos(M_PI*a/2)+pow(v,2*a), -g/2 ) *
        cos( g*aux_phi_havneg( v, a ) ); }

double func_im_havneg (double v, double a, double g) {
    return pow( 1 + 2*pow(v,a)*cos(M_PI*a/2)+pow(v,2*a), -g/2 ) *
        sin( g*aux_phi_havneg( v, a ) ); }


//**************************************************************************************************
//*  Functions of three arguments
//**************************************************************************************************

int func_cond(int v, int a, int b) { return v ? a : b; }
double func_cond(double v, double a, double b) { return v ? a : b; }
void func_cond(double& r, double& dr, double v, double dv, double a, double da, double b, double db)
{ if(v) { r=a; dr=da; } else { r=b; dr=db; } }

double func_q4w( double hbarw, double Ei, double twotheta) {
    // calculate Q in Angstroem-1 when hbar*w and E_i are given in meV
    if( hbarw>Ei ) return -1;
    return sqrt( (1/2.07212349) *
                 ( 2*Ei-hbarw-2*sqrt(Ei*(Ei-hbarw))*cos(twotheta*TWOPI/360) ) );
}

double func_cauchy2 (double x, double p, double w) {
    return w==0 ? 0 : w/PI*(1.0/(SQR(x-p)+SQR(w))+1.0/(SQR(x+p)+SQR(w)));
}

double func_kwwc( double v, double t, double b ){
    if( b<0.1 || b>2.0 ) return 0;
    return t/PI*kwwc( v*t, b ); }
double func_kwws( double v, double t, double b ){
    if( b<0.1 || b>2.0 ) return 0;
    return t/PI*kwws( v*t, b ); }
double func_kwwp( double v, double t, double b ){
    if( b<0.1 || b>2.0 || t<=0 ) return 0;
    return 1/PI*kwwp( v*t, b ); }

double func_kwwmc( double v, double tmean, double b ){
    if( b<0.1 || b>2.0 ) return 0;
    double tkww = tmean*b/tgamma(1/b);
    return tkww/PI*kwwc( v*tkww, b ); }
double func_kwwms( double v, double tmean, double b ){
    if( b<0.1 || b>2.0 ) return 0;
    double tkww = tmean*b/tgamma(1/b);
    return tkww/PI*kwws( v*tkww, b ); }
double func_kwwmp( double v, double tmean, double b ){
    if( b<0.1 || b>2.0 || tmean<=0 ) return 0;
    double tkww = tmean*b/tgamma(1/b);
    return 1/PI*kwwp( v*tkww, b ); }

double func_voigt( double v, double s, double g ){
    return (s||g) ? voigt(v,s,g) : 0;
}

double func_zorn( double v, double vavg, double s ) {
    if( s<=0 || s>=1 ) return 0;
    return (1-s)*(v+s*vavg*vavg/(1-s*vavg)); }

double func_zorn_gauss( double q, double u2, double s ) {
    if( s<=0 || s>=1 ) return 0;
    double qmax = 4*PI/6.27; // Si111
    double v = exp(-u2*SQR(q));
    double vavg = (1-exp(-u2*SQR(qmax))) / (u2*SQR(qmax));
    return (1-s)*(v+s*vavg*vavg/(1-s*vavg)); }

double func_rrdm( double wred, double act0, double relwidth ) {
    // Rotational rate distribution model [review: Colmenero 2005]
    //   Gaussian distribution of Barriers, Arrhenius law, rotational jumps.
    // Dimensionless:
    //   wred = w/Omega_0,
    //   act0 = mean{E_A}/(k_B T),
    //   relwidth = stdv{E_A} / mean{E_A};
    //   multiply result by T/Omega_0.
    const int N=101;
    double actsig = act0*relwidth;
    double actrange = 3*actsig;
    if ( act0-actrange <= 0 )
        return 0;
    double da = actrange/((N-1)/2);
    double ret = 0;
    for( int i=0; i<N; ++i ) {
        double a = act0 + da*(i-(N-1)/2);
        ret += exp( -SQR((a-act0)/actsig)/2 ) *func_cauchy( wred, exp(-a) );
    }
    return ret * da / sqrt(2*PI) / actsig;
}

double func_rotdiff( double w, double tau, double qb )
// Rotational diffusion spectrum
{
    double ret = 0;
    if( qb<=0 )
        return 0;
    for( int l=1; l<=12; ++l ){
        ret += (2*l+1) * pow( gsl_sf_bessel_jl( l, qb ), 2 ) *
            l*(l+1)*tau/PI/(SQR(w*tau)+SQR(l*(l+1)));
    }
    return ret;
}

double func_ornuhl( double _w, double _a, double _b )
// Ornstein-Uhlenbeck spectrum FT[exp(-|b|t-a(1-exp(-t))),
//    computed for w->0 by expanding the outer exponential in e^(a e^(-x/w))
//    which turns out to converge rapidly for all w
{
    const double ornuhl_delta=2.2e-16, ornuhl_eps=5.5e-20;
    long double w = _w;
    long double a = _a;
    long double b = fabs(_b);
    if ( a==0 )
        return b/(SQR(b)+SQR(w));
    long double a_log = logl( a );
    long double S=0;      // summed series
    long double h;        // exponent
    long double f;        // Cauchy-Lorentz function
    long double u;        // precomputed common factors
    long double u_next=0; // - next value [initialized to avoid warning]

    // sum the expansion
    for ( int k=0; k<1000; ++k ) {
        u = u_next;
        h = k*a_log - lgammal((long double)(k+1));
        if ( h>DBL_MAX_EXP )
            return -3; // exponent overflow
        if ( h<DBL_MIN_EXP )
            return -4; // exponent underflow
        f = (k+b)/(SQR(k+b)+SQR(w));
        u_next = expl( h ) *f;
        S += u;
        if( k==0 )
            continue;
        // termination criteria
        if ( ornuhl_eps*S+u_next <= ornuhl_delta*S )
            return exp(-a)*S; // reached required precision
        else if ( u_next<DBL_MIN )
            return -7; // underflow
    }
    return -9; // too many terms
}

//**************************************************************************************************
//*  Registration
//**************************************************************************************************

//! Registers basic and arithmetic functions.

void fbase_initialize()
{
    auto G = SFuncRegistry::instance();
    CFuncMetadata m;

    // operators by precedence (which should agree with xax_yacc.ypp):
    m = { "+-",  2, "a+-da   [set error]", 1  };
    G->register_fct_e_dd ( m, func_pm );
    m = { "^",   2, "a^b     [exponentiation]", 2 };
    G->register_fct_d_dd ( m, func_pow );
    G->register_fct_e_ee ( m, func_pow );
    m = { "!",   1, "!a      [negation]", 3 };
    G->register_fct_i_i  ( m, func_not );
    G->register_fct_i_d  ( m, func_not );
    m = { "neg", 1, "-a      [negative value]", 4 };
    G->register_fct_i_i  ( m, func_neg );
    G->register_fct_d_d  ( m, func_neg );
    G->register_fct_e_e  ( m, func_neg );
    m = { "/",   2, "a/b     [quotient]", 5 };
    G->register_fct_d_dd ( m, func_div );
    G->register_fct_e_ee ( m, func_div );
    m = { "//",  2, "i//j    [integer quotient]", 5 };
    G->register_fct_i_ii ( m, func_idiv );
    G->register_fct_i_di ( m, func_idiv );
    m = { "%",   2, "i%j     [remainder]", 5 };
    G->register_fct_i_ii ( m, func_mod );
    G->register_fct_d_dd ( m, func_mod );
    m = { "*",   2, "a*b     [product]", 6 };
    G->register_fct_i_ii ( m, func_mul );
    G->register_fct_d_dd ( m, func_mul );
    G->register_fct_e_ee ( m, func_mul );
    m = { "+",   2, "a+b     [sum]", 7 };
    G->register_fct_i_ii ( m, func_add );
    G->register_fct_d_dd ( m, func_add );
    G->register_fct_e_ee ( m, func_add );
    m = { "-",   2, "a-b     [difference]", 7 };
    G->register_fct_i_ii ( m, func_sub );
    G->register_fct_d_dd ( m, func_sub );
    G->register_fct_e_ee ( m, func_sub );
    m = { ">",   2, "a>b     [greater]", 8 };
    G->register_fct_i_ii ( m, func_gt );
    G->register_fct_i_dd ( m, func_gt );
    m = { "<",   2, "a<b     [less]", 8 };
    G->register_fct_i_ii ( m, func_lt );
    G->register_fct_i_dd ( m, func_lt );
    m = { ">=",  2, "a>=b    [greater or equal]", 8 };
    G->register_fct_i_ii ( m, func_ge );
    G->register_fct_i_dd ( m, func_ge );
    m = { "<=",  2, "a<=b    [less or equal]", 8 };
    G->register_fct_i_ii ( m, func_le );
    G->register_fct_i_dd ( m, func_le );
    m = { "==",  2, "a==b    [equality]", 9 };
    G->register_fct_i_ii ( m, func_eq );
    G->register_fct_i_dd ( m, func_eq );
    m = { "!=",  2, "a!=b    [unequality]", 9 };
    G->register_fct_i_ii ( m, func_ne );
    G->register_fct_i_dd ( m, func_ne );
    m = { "xor", 2, "a^b     [exclusive or]", 10 };
    G->register_fct_i_ii ( m, func_xor );
    G->register_fct_i_dd ( m, func_xor );
    m = { "&&",  2, "a&&b    [and]", 11 };
    G->register_fct_i_ii ( m, func_and );
    G->register_fct_i_dd ( m, func_and );
    m = { "||",  2, "a||b    [or]", 11 };
    G->register_fct_i_ii ( m, func_or );
    G->register_fct_i_dd ( m, func_or );
    m = { "?:",  2, "a?b:c   [if a then b else c]", 13 };
    G->register_fct_i_iii( m, func_cond );
    G->register_fct_d_ddd( m, func_cond );
    G->register_fct_e_eee( m, func_cond );


    // f( 1 arg )
    m = { "to_d", 1, "(x): convert x to floating-point number" };
    G->register_fct_d_i  ( m, func_to_d );
    G->register_fct_d_d  ( m, func_to_d );
    G->register_fct_d_s  ( m, func_to_d );
    m = { "pop", 1, "(x): do nothing" };
    G->register_fct_0_i  ( m, func_pop );
    G->register_fct_0_d  ( m, func_pop );
    G->register_fct_0_s  ( m, func_pop );
    m = { "exit", 1, "(x): exit(x)" };
    G->register_fct_i_i  ( m, func_exit );
    m = { "ln", 1, "(x): natural logarithm of x, or 0 if x<=0" };
    G->register_fct_d_d  ( m, func_ln );
    G->register_fct_e_e  ( m, func_ln );
    m = { "lg", 1, "(x): decadic logarithm of x, or 0 if x<=0" };
    G->register_fct_d_d  ( m, func_lg );
    G->register_fct_e_e  ( m, func_lg );
    m = { "sqrt", 1, "(x): square root of x, or 0 if x<0" };
    G->register_fct_d_d  ( m, func_sqrt );
    G->register_fct_e_e  ( m, func_sqrt );
    m = { "abs", 1, "(x): absolute value of x" };
    G->register_fct_i_i  ( m, func_abs );
    G->register_fct_d_d  ( m, func_abs );
    G->register_fct_e_e  ( m, func_abs );
    m = { "sign", 1, "(x): sign of x" };
    G->register_fct_i_i  ( m, func_sign );
    G->register_fct_d_d  ( m, func_sign );
    G->register_fct_e_e  ( m, func_sign );
    m = { "exp", 1, "(x): exponential function of x" };
    G->register_fct_d_d  ( m, func_exp );
    G->register_fct_e_e  ( m, func_exp );
    m = { "sin", 1, "(x): sine of x (where x in radian)" };
    G->register_fct_d_d  ( m, func_sin );
    G->register_fct_e_e  ( m, func_sin );
    m = { "cos", 1, "(x): cosine of x (where x in radian)" };
    G->register_fct_d_d  ( m, func_cos );
    G->register_fct_e_e  ( m, func_cos );
    m = { "tan", 1, "(x): tangent of x (where x in radian)" };
    G->register_fct_d_d  ( m, func_tan );
    m = { "cot", 1, "(x): cotangent of x (where x in radian)" };
    G->register_fct_d_d  ( m, func_cot );
    m = { "sind", 1, "(x): sine of x (where x in degrees)" };
    G->register_fct_d_d  ( m, func_sind );
    m = { "cosd", 1, "(x): cosine of x (where x in degrees)" };
    G->register_fct_d_d  ( m, func_cosd );
    m = { "tand", 1, "(x): tangent of x (where x in degrees)" };
    G->register_fct_d_d  ( m, func_tand );
    m = { "cotd", 1, "(x): cotangent of x (where x in degrees)" };
    G->register_fct_d_d  ( m, func_cotd );
    m = { "asin", 1, "(x): arc sine of x (result is in radian; 0 if |x|>1)" };
    G->register_fct_d_d  ( m, func_asin );
    m = { "acos", 1, "(x): arc cosine of x (result is in radian; 0 if |x|>1)" };
    G->register_fct_d_d  ( m, func_acos );
    m = { "atan", 1, "(x): arc tangent of x (result is in radian)" };
    G->register_fct_d_d  ( m, func_atan );
    m = { "acot", 1, "(x): arc cotangent of x (result is in radian)" };
    G->register_fct_d_d  ( m, func_acot );
    m = { "asind", 1, "(x): arc sine of x (result is in degrees; 0 if |x|>1)" };
    G->register_fct_d_d  ( m, func_asind );
    m = { "acosd", 1, "(x): arc cosine of x (result is in degrees; 0 if |x|>1)" };
    G->register_fct_d_d  ( m, func_acosd );
    m = { "atand", 1, "(x): arc tangent of x (result is in degrees)" };
    G->register_fct_d_d  ( m, func_atand );
    m = { "acotd", 1, "(x): arc cotangent of x (result is in degrees)" };
    G->register_fct_d_d  ( m, func_acotd );
    m = { "sinh", 1, "(x): hyperbolic sine of x" };
    G->register_fct_d_d  ( m, func_sinh );
    m = { "cosh", 1, "(x): hyperbolic cosine of x" };
    G->register_fct_d_d  ( m, func_cosh );
    m = { "tanh", 1, "(x): hyperbolic tangent of x" };
    G->register_fct_d_d  ( m, func_tanh );
    m = { "coth", 1, "(x): hyperbolic tangent of x" };
    G->register_fct_d_d  ( m, func_coth );
    m = { "j0", 1, "(x): spherical Bessel function x" };
    G->register_fct_d_d  ( m, func_j0 );
    m = { "j1", 1, "(x): spherical Bessel function x" };
    G->register_fct_d_d  ( m, func_j1 );
    m = { "fac", 1, "(x): factorial of nearest integer of x" };
    G->register_fct_d_d  ( m, func_fac );
    m = { "cata", 1, "(x): Catalan number of nearest integer of x" };
    G->register_fct_d_d  ( m, func_cata );
    m = { "gamma", 1, "(x): gamma function of x (i.e. factorial of x-1)" };
    G->register_fct_d_d  ( m, func_gamma );
    m = { "lgamma", 1, "(x): ln of gamma of x" };
    G->register_fct_d_d  ( m, func_lgamma );
    m = { "debye1", 1, "(x): Debye function D1 of x" };
    G->register_fct_d_d  ( m, func_debye1 );
    m = { "debye3", 1, "(x): Debye function D3 of x" };
    G->register_fct_d_d  ( m, func_debye3 );
    m = { "debyeu2", 1, "(x): Debye mean squared displacement's temperature dependence" };
    G->register_fct_d_d  ( m, func_debye_msd );
    m = { "debyeui", 1, "(x): Debye internal energy" };
    G->register_fct_d_d  ( m, func_debye_ui );
    m = { "debyecv", 1, "(x): Debye heat capacity" };
    G->register_fct_d_d  ( m, func_debye_cv );
    m = { "erfP", 1, "(x): ?" };
    G->register_fct_d_d  ( m, func_erfP );
    m = { "erfQ", 1, "(x): ?" };
    G->register_fct_d_d  ( m, func_erfQ );
    m = { "erf", 1, "(x): error function of x" };
    G->register_fct_d_d  ( m, func_erf );
    m = { "erfc", 1, "(x): complementary error function of x" };
    G->register_fct_d_d  ( m, func_erfc );
    m = { "erfi", 1, "(x): imaginary error function of x" };
    G->register_fct_d_d  ( m, func_erfi );
    m = { "dawson", 1, "(x): Dawson function of x" };
    G->register_fct_d_d  ( m, func_dawson );
    m = { "sinc", 1, "(x): sinus cardinalis, sin(x)/x" };
    G->register_fct_d_d  ( m, func_sinc );
    m = { "ceil", 1, "(x): the smallest integer i with i>=x" };
    G->register_fct_i_d  ( m, func_ceil );
    m = { "floor", 1, "(x): the largest integer i with i<=x" };
    G->register_fct_i_d  ( m, func_floor );
    m = { "nint", 1, "(x): the integer nearest to x" };
    G->register_fct_i_d  ( m, func_nint );

    m = { "diehl", 1, "(cauchywid/gausswid): normalized convolution gauss(*)cauchy" };
    G->register_fct_d_d  ( m, func_diehl );
    m = { "lndiehl", 1, "(cauchywid/gausswid): ln of normalized convolution gauss(*)cauchy" };
    G->register_fct_d_d  ( m, func_lndiehl );

    // f(2 args)
    m = { "exit_if", 2, "(x,s): throw s if x" };
    G->register_fct_i_is ( m, func_exit_if );
    m = { "exit_unless", 1, "(x,s): throw s if !x" };
    G->register_fct_i_is ( m, func_exit_unless );
    m = { "min2", 2, "(x,y): the smaller of the two arguments x and y" };
    G->register_fct_i_ii ( m, func_min );
    G->register_fct_d_dd ( m, func_min );
    m = { "max2", 2, "(x,y): the smaller of the two arguments x and y" };
    G->register_fct_i_ii ( m, func_max );
    G->register_fct_d_dd ( m, func_max );
    m = { "ran", 2, "(x,y): a random number between x and y" };
    G->register_fct_d_dd ( m, func_ran );

    m = { "gauss", 2, "(x,s): the normalized Gaussian exp(-x^2/2/s^2)/sqrt(2 pi)/s" };
    G->register_fct_d_dd ( m, func_gauss );
    m = { "gnn", 2, "(x,s): the unnormalized Gaussian exp(-x^2/2/s^2)" };
    G->register_fct_d_dd ( m, func_gaussnn );
    m = { "cauchy", 2, "(x,w): the Cauchy-Lorentz function ?/(x^2+w^2)" };
    G->register_fct_d_dd ( m, func_cauchy );

    // f(3 args)
    m = { "rehavneg", 3, "(x,y,z): real part of the Havriliak-Negami function" };
    G->register_fct_d_ddd( m, func_re_havneg );
    m = { "imhavneg", 3, "(x,y,z): imaginary part of the Havriliak-Negami function" };
    G->register_fct_d_ddd( m, func_im_havneg );
    m = { "q4w", 3, "(x,y,z): ?" };
    G->register_fct_d_ddd( m, func_q4w );
    m = { "cauchy2", 3, "(x,y,z): ?" };
    G->register_fct_d_ddd( m, func_cauchy2 );
    m = { "kwwc", 3, "(w,tau,b): Fourier cosine transform (t->w) of exp((t/tau)^b)" };
    G->register_fct_d_ddd( m, func_kwwc );
    m = { "kwws", 3, "(w,tau,b): Fourier sine transform (t->w) of exp((t/tau)^b)" };
    G->register_fct_d_ddd( m, func_kwws );
    m = { "kwwp", 3, "(w,tau,b): primitive of Fourier cosine transform (t->w) of exp((t/tau)^b)" };
    G->register_fct_d_ddd( m, func_kwwp );
    m = { "kwwmc", 3, "(w,<tau>,b): Fourier cosine transform (t->w) of exp((t/tau)^b)" };
    G->register_fct_d_ddd( m, func_kwwc );
    m = { "kwwms", 3, "(w,<tau>,b): Fourier sine transform (t->w) of exp((t/tau)^b)" };
    G->register_fct_d_ddd( m, func_kwws );
    m = { "kwwmp", 3, "(w,<tau>,b): primitive of Fourier cosine transform (t->w) of exp((t/tau)^b)" };
    G->register_fct_d_ddd( m, func_kwwp );
    m = { "voigt", 3, "(x,sigma,gamma): convolution of Gaussian(x,sigma) and Lorentzian(x,gamma)" };
    G->register_fct_d_ddd( m, func_voigt );
    m = { "zorn", 3, "(I,<I>,s): Zorn's multiple-scattering corrected elastic intensity" };
    G->register_fct_d_ddd( m, func_zorn );
    m = { "zorn2", 3, "(q,<u^2>,s): Zorn's multiple-scattering corrected Gaussian elastic intensity for Si111" };
    G->register_fct_d_ddd( m, func_zorn_gauss );
    m = { "rrdm", 3, "(w*t0,EA_mean/T,EA_stdv/EA_mean: rotational rate distribution model" };
    G->register_fct_d_ddd( m, func_rrdm );
    m = { "ornuhl", 3, "(w,a,b) Ornstein-Uhlenbeck spectrum FT[exp(-|b|t-a(1-exp(-t)))]" };
    G->register_fct_d_ddd( m, func_ornuhl );
    m = { "rotdiff", 3, "(w,tau,qb: rotational diffusion spectrum)" };
    G->register_fct_d_ddd( m, func_rotdiff );
}
