//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  file_in.cpp
//! \brief NFileIn: load files.

#include "defs.hpp"

#include <fstream>
#include <string.h>

#include <yaml-cpp/yaml.h>

#include "../trivia/string_ops.hpp"
#include "../trivia/file_ops.hpp"
#include "../readplus/ask.hpp"

#include "olf.hpp"
#include "mem.hpp"
#include "slice.hpp"
#include "obj.hpp"
#include "file_in.hpp"


namespace NFileIn {
    void Load_yda( std::ifstream& F_in, string fnam );
    void Load_08( std::ifstream& F_in, string flong );
    void Load_01( FILE *F_in, string flong );
    void LoadSpec_01( FILE *F_in, PSpec& sout, int nz );
    void Load_96( FILE *F_in, string flong );
}


//! Load file in current or legacy format.

void NFileIn::load(void)
{
    string pattern = sask( "Load file(s)" );
    vector<string> fNames = triv::glob_file_list( pattern, "yda y08 a01 i96" );
    for( string fnam: fNames ) {
        try{
            string fdir, fshort, fext;
            triv::fname_divide( fnam, &fdir, &fshort, &fext);
            if ( fext=="y08" || fext=="yda" ) {
                std::ifstream FS (fnam.c_str(), std::ios_base::in);
                if( FS.fail() )
                    throw string("cannot open file");
                cout << ".. loading file " << fnam << "\n";
                Load_yda( FS, fshort );
                FS.close();
            } else if ( fext=="a01" || fext=="i96" ) {
                FILE *F_in;
                if( !(F_in = fopen(fnam.c_str(), "r")) )
                    throw "cannot open file " + fnam;
                if ( fext=="a01" ) {
                    Load_01( F_in, fnam.c_str() );
                } else if ( fext=="i96" ) {
                    Load_96( F_in, fnam.c_str() );
                }
                fclose( F_in );
            } else
                throw "unknown extension " + fext;
        }
        catch(string& ex) {
            throw "Cannot load " + fnam + ": " + ex;
        }
        catch(std::exception& ex) {
            throw "Cannot load " + fnam + ": " + ex.what();
        }
    }
}

//! Load a YAML file in yda format.

void NFileIn::Load_yda(std::ifstream& FS, string fnam)
{
    const YAML::Node doc = YAML::Load(FS);
    if( doc.Type() != YAML::NodeType::Map )
        throw S("document root is not of MAP type");

    //read meta
    if(!doc["Meta"])
        throw S("no Meta");
    if ( doc["Meta"].Type() != YAML::NodeType::Map )
        throw S("Meta is not of MAP type");

    if(!doc["Meta"]["format"])
        throw S("no format in Meta");
    string s = doc["Meta"]["format"].as<string>();
    int version;
    if      ( s == "frida/y08 for yaml1" )
        version = 1;
    else if ( s == "yaml/frida 2.0" )
        version = 2;
    else
        throw "unknown data format '" + s + "'";

    if(!doc["Meta"]["type"])
        throw S("no type in Meta");
    string type = doc["Meta"]["type"].as<string>();

    // create output file
    POlo fout;
    bool isdata;
    COlc *fc = nullptr;
    if ( type == "generic tabular data" ){
        isdata = true;
        fout = POld( new COld );
    } else if ( type == "frida2 curve" ){
        isdata = false;
        fout = POlc( new COlc );
        fc = dynamic_cast<COlc*>(fout.get());
     } else
        throw "Invalid file type "+type;

    // read history
    if(!doc["History"])
        throw S("no History");
    if ( doc["History"].Type() != YAML::NodeType::Sequence &&
         (version>1 || doc["History"].size()>0 ) )
        throw S("History is not of SEQUENCE type");
    for (int i = 0; i < doc["History"].size(); i++) {
        string s = doc["History"][i].as<string>();
        fout->log_action(s);
    }

    // read coord
    if(!doc["Coord"])
        throw S("no Coord");
    if ( doc["Coord"].Type() != YAML::NodeType::Map )
        throw S("Coord is not of MAP type");
    if( const YAML::Node& dCo = doc["Coord"]["x"] ){
        fout->xco.name = dCo["name"].as<string>();
        fout->xco.unit =dCo["unit"].as<string>();
    } else
        throw S("no x coord");
    if( const YAML::Node& dCo = doc["Coord"]["y"] ){
        fout->yco.name = dCo["name"].as<string>();
        fout->yco.unit = dCo["unit"].as<string>();
    } else
        throw S("no y coord");
    if ( version==1 ) {
        int iz = 0;
        while(const YAML::Node& dCo = doc["Coord"]["z"+S(iz)]){
            CCoord co(dCo["name"].as<string>(), dCo["unit"].as<string>());
            fout->ZCo.push_back( co );
            ++iz;
        }
    } else {
        if ( doc["Coord"]["z"].Type() != YAML::NodeType::Sequence )
            throw S("Coord[z] is not of SEQUENCE type");
        for ( auto dCo: doc["Coord"]["z"] ) {
            CCoord co(dCo["name"].as<string>(), dCo["unit"].as<string>());
            fout->ZCo.push_back( co );
        }
    }

    // read r-pars
    if ( version==1 ) {
        if(!doc["Param"])
            throw S("no Param");
        if ( doc["Param"].Type() != YAML::NodeType::Sequence  && doc["Param"].size() > 0)
            throw S("Param is not of SEQUENCE type");
        for( int iParam = 0; iParam < (doc["Param"].size()); iParam++ ){
            if ( doc["Param"][iParam].Type() != YAML::NodeType::Map )
                throw "Param " + S(iParam) + " is not of MAP type";
            CCoord co(doc["Param"][iParam]["name"].as<string>(),
                      doc["Param"][iParam]["unit"].as<string>());
            string s = doc["Param"][iParam]["value"].as<string>();
            double num;
            if( !triv::any2dbl( s, &num ) )
                throw "param(" + co.name + "): invalid value " + s;
            fout->RPar.push_back( CParam( co, num ) );
        }
    } else {
        if(!doc["RPar"])
            throw S("no RPar");
        if ( doc["RPar"].Type() != YAML::NodeType::Sequence )
            throw S("RPar is not of SEQUENCE type");
        for ( auto dCo: doc["RPar"] ) {
            if ( dCo.Type() != YAML::NodeType::Map )
                throw S("an r-par is not of MAP type");
            CCoord co(dCo["name"].as<string>(), dCo["unit"].as<string>());
            double val = dCo["val"].as<double>();
            fout->RPar.push_back( CParam( co, val ) );
        }
    }

    // for curve, read formula
    if( !isdata ) {
       string expr;
        if ( version==1 ) {
            if( !doc["Formula"] )
                throw S("DEFICIENT FILE: no formula");
            expr = doc["Formula"].as<string>();
        } else {
            const YAML::Node& nod = doc["Curve"];
            if( !nod )
                throw S("DEFICIENT FILE: no Curve section");
            expr = nod["expr"].as<string>();
            fc->nP = nod["nPar"].as<int>();
            fc->weighing = COlc::name2wgt( nod["weighing"].as<string>() );
            fc->chi2 = nod["chi2"].as<double>();
        }
        fc->parse_function( expr );
        fc->curve_set_defaults();
    }

    // start to read slices
    if ( version==1 ) {
        if(!doc["Tables"])
            throw S("no Tables");
        if ( doc["Tables"].Type() != YAML::NodeType::Sequence && doc["Tables"].size() > 0 )
            throw S("Tables is not of SEQUENCE type");
        for (int iTable = 0; iTable < (doc["Tables"].size()); iTable++) {
            if ( doc["Tables"][iTable].Type() != YAML::NodeType::Map )
                throw "Tables " + S(iTable) + " is not of MAP type";
            vector<RObj> z;
            for(int iz=0; iz<fout->ZCo.size(); ++iz ){
                string s = doc["Tables"][iTable]["z" + S(iz)].as<string>() ;
                double num;
                if( !triv::any2dbl( s, &num ) )
                    throw "z" + S(iz) + ": invalid value " + s;
                z.push_back( RObjDbl( new CObjDbl( num ) ) );
            }

            if( isdata ) {
                PSpec sout( new CSpec );
                sout->z = z;
                string xytag;
                if ( doc["Tables"][iTable]["xy"] ) {
                    string res, lin;
                    string val = doc["Tables"][iTable]["xy"].as<string>();
                    while( val != "" ) {
                        triv::string_extract_line( val, &lin, &res );
                        double vx, vy;
                        if( sscanf( lin.c_str(), "%lg %lg\n", &vx, &vy )!=2 )
                            throw "xy: bad data line: " + lin;
                        sout->push_xy( vx, vy );
                        val = res;
                    };
                } else if ( doc["Tables"][iTable]["xyd"] ) {
                    //itxy.second() >> val;
                    string res, lin;
                    string val = doc["Tables"][iTable]["xyd"].as<string>();
                    while( val != "" ) {
                        triv::string_extract_line( val, &lin, &res );
                        double vx, vy, dy;
                        if( sscanf( lin.c_str(), "%lg %lg %lg\n",&vx, &vy, &dy )!=3 )
                            throw "xyd: bad data line: " + lin;
                        sout->push_xyd( vx, vy, dy );
                        val = res;
                    };
                } else
                    throw "invalid xytag " + xytag;
                fout->V.push_back( move(sout) );
            } else {
                PCurve cout( new CCurve );
                cout->z = z;
                for( int ip=0; ip<fc->nP; ++ip ){
                    string res;
                    string val = doc["Tables"][iTable]["p" + S(ip)].as<string>();
                    double num;
                    if( !triv::any2dbl( val, &num ) )
                        throw "p" + S(ip) + ": invalid value " + val;
                    cout->P.push_back( num );
                    cout->ParAttr.push_back( 'u' );
                }
                fout->V.push_back( move(cout) );
            }
        }
    } else {
        if(!doc["Slices"])
            throw S("no Slices");
        if ( doc["Slices"].Type() != YAML::NodeType::Sequence )
            throw S("section Slices is not of SEQUENCE type");
        for ( const YAML::Node& nodeSlic: doc["Slices"] ) {
            if ( nodeSlic.Type() != YAML::NodeType::Map )
                throw S("a slice is not of MAP type");
            const YAML::Node& nodeZ = nodeSlic["z"];
            if(!nodeZ)
                throw S("a slice has no z entry");
            if ( nodeZ.Type() != YAML::NodeType::Sequence )
                throw S("a z entry is not of SEQUENCE type");
            vector<RObj> z;
            for ( const YAML::Node& nodeZi: nodeZ ) {
                if ( nodeSlic.Type() != YAML::NodeType::Map )
                    throw S("a z entry is not of MAP type");
                z.push_back( RObjDbl( new CObjDbl( nodeZi["val"].as<double>() ) ) );
            }
            if( isdata ) {
                PSpec sout( new CSpec );
                sout->z = z;
                if(!nodeSlic["x"])
                    throw S("a slice has no x entry");
                if ( nodeSlic["x"].Type() != YAML::NodeType::Sequence )
                    throw S("an x entry is not of SEQUENCE type");
                for ( const YAML::Node& nodeEle: nodeSlic["x"] )
                    sout->x.push_back( nodeEle.as<double>() );

                if(!nodeSlic["y"])
                    throw S("a slice has no y entry");
                if ( nodeSlic["y"].Type() != YAML::NodeType::Sequence )
                    throw S("a y entry is not of SEQUENCE type");
                for ( const YAML::Node& nodeEle: nodeSlic["y"] )
                    sout->y.push_back( nodeEle.as<double>() );
                if( sout->y.size() != sout->x.size() )
                    throw S("a y entry differs in length from x");

                if( nodeSlic["dy"] ) {
                    if ( nodeSlic["dy"].Type() != YAML::NodeType::Sequence )
                        throw S("a dy entry is not of SEQUENCE type");
                    for ( const YAML::Node& nodeEle: nodeSlic["dy"] )
                        sout->dy.push_back( nodeEle.as<double>() );
                    if( sout->dy.size() != sout->x.size() )
                        throw S("a dy entry differs in length from x");
                }
                fout->V.push_back( move(sout) );
            } else {
                PCurve cout( new CCurve );
                cout->z = z;
                if(!nodeSlic["p"])
                    throw S("a slice has no p entry");
                if ( nodeSlic["p"].Type() != YAML::NodeType::Sequence )
                    throw S("a p entry is not of SEQUENCE type");
                for ( const YAML::Node& nodeEle: nodeSlic["p"] )
                    cout->P.push_back( nodeEle.as<double>() );
                if ( cout->P.size()!=fc->nP )
                    throw S("number of p entries conflicts with nPar");

                if(!nodeSlic["attr"])
                    throw S("a slice has no attr entry");
                if ( nodeSlic["attr"].Type() != YAML::NodeType::Sequence )
                    throw S("an attr entry is not of SEQUENCE type");
                for ( const YAML::Node& nodeEle: nodeSlic["attr"] )
                    cout->ParAttr.push_back( nodeEle.as<char>() );
                if ( cout->ParAttr.size()!=fc->nP )
                    throw S("number of attr entries conflicts with nPar");

                cout->fitOutcome = nodeSlic["fitOutcome"].as<double>();
                cout->fitChi2 = nodeSlic["fitChi2"].as<double>();
                cout->fitR2 = nodeSlic["fitR2"].as<double>();

                fout->V.push_back( move(cout) );
            }
        }
    }

    fout->name = fnam;
    fout->as_on_disk = true;
    SMem::instance()->mem_store( move(fout) );
}


//! Load a .a01 formatted file.

void NFileIn::Load_01( FILE *F_in, string flong )
{
    string lin, key, val, fdir, fshort, fext;
    int iz, nj=0, j;
    CCoord co;
    POld fout( new COld );

    triv::fname_divide( flong, &fdir, &fshort, &fext);

    if     ( triv::freadln(F_in, &lin)<=0 )
        throw  "file empty or starting with empty line";
    if     ( lin!=string("ASCII-01 JWu") )
        throw ".a01 file starts with illegal header [" + lin + "]";

    if ( triv::freadln(F_in, &lin)<=0 || lin.length()<5 ||
         lin.substr(0,4)!=string("typ ") )
        throw "no valid typ line";
    if        ( lin.substr(4)=="D" ) {
        // ok
    } else if ( lin.substr(4)=="C" ) {
        throw "reading curve files not implemented";
    } else {
        throw "invalid type [" + lin.substr(4) + "]";
    }

    while (1) { // loop over input lines
        if ( !triv::freadln(F_in, &lin) )
            throw "unexpected end-of-file";

        // regular exit when "eof" is reached:
        if ( lin=="eof" ) {
            fout->as_on_disk = true;
            SMem::instance()->mem_store( move(fout) );
            return;
        }

        if ( lin.size() < 4 )
            throw "incomplete line";

        key = lin.substr(0,4);
        val = lin.substr(4);

        if        (key=="fil ") {
            if (val!=fshort) {
                cout << "warning: file '" << fshort <<
                    "' renamed from '" << val << "'\n";
            }
            fout->name = fshort;
        } else if (key=="doc ") {
            fout->lDoc.push_back(val);
        } else if (key=="x   ") {
            fout->xco.load_a01(val);
        } else if (key=="y   ") {
            fout->yco.load_a01(val);
        } else if (lin[0]=='z') {
            if ( sscanf(lin.substr(1,3).c_str(), "%i", &iz)!=1 )
                throw "z coordinates must be numbered";
            if ( iz!=fout->ZCo.size() )
                throw "z coordinates must be sorted";
            co.load_a01(val);
            fout->ZCo.push_back(co);
        } else if (lin.substr(0,4)=="rpa ") {
            string in = lin.substr(4);
            int ji = in.find_first_not_of(" \t");
            double dval;
            if( ji>=20 || !triv::any2dbl( in.substr(ji,20), &dval ) )
                throw "invalid rpa line\n";
            fout->RPar.push_back( CParam( CCoord(), dval ) );
            throw "a01 file contains rpa section. "
                "Check the source code preceeding this error message. "
                "It is possibly broken. If not, remove this exception.";
        } else if (key=="#j  ") {
            if ( sscanf(val.c_str(), "%i", &nj)!=1 )
               throw "after key #j no no. of spectra";
        } else if (key=="scn ") {
            if ( sscanf(val.c_str(), "%i", &j)!=1 ) {
                cout << "warning: number of spectra not given\n";
            } else if ( j!=fout->nJ() ) {
                cout << "warning: spec " << fout->nJ() <<
                    " has label " << j << "\n";
            }
            PSpec sout( new CSpec );
            LoadSpec_01( F_in, sout, fout->nZ() );
            fout->V.push_back( move(sout) );
        } else {
            throw "Unexpected key [" + key + "]";
        }
    }
}


//! Load a spectrum from a .a01 formatted files.

void NFileIn::LoadSpec_01( FILE *F_in, PSpec& sout, int nz )
{
    int err;
    int i, m, n;
    char lab[4];
    double v, w;

    // get z values:
    if ((err=fscanf(F_in, "%4c%i\n", lab, &m))!=2)
        throw "expecting Z head line, found " + S(err) + " entries";
    if (strncmp(lab, "Z   ", 4))
        throw "found [" + string(lab) + "] in Z head line";
    if (m!=nz)
        throw "found " + S(m) + " Z entries, expecting " + S(nz);
    for ( i=0; i<nz; ++i ) {
        if ((err=fscanf(F_in, "%lg\n", &v))!=1)
            throw "expecting z value, found " + S(err) + " entries";
        sout->z.push_back( PObjDbl( new CObjDbl( v ) ) );
    }

    // get xy values:
    if ( (err=fscanf(F_in, "%4c%i\n", lab, &n))!=2 )
        throw "expecting XY head line, found " + S(err) + " entries";
    if (strncmp(lab, "XY  ", 4) )
        throw "found [" + string(lab) + "] in XY head line";

    for (i=0; i<n; i++) {
        if ((err=fscanf(F_in, "%lg %lg\n", &v, &w))!=2) {
            throw "expecting xy pair, found S(err) entries";
        }
        sout->push_xy(v, w);
    }
}


//! Load a .i96 formatted file.

void NFileIn::Load_96( FILE *F_in, string flong )
{
    string lin, key, val, fdir, fshort, fext;
    double r0, r1, r2;
    int n, nspec, ival;
    POld fout( new COld );
    CCoord co;
    PSpec sout;

    triv::fname_divide( flong, &fdir, &fshort, &fext);
    fout->name = fshort;

    if     ( !triv::freadln(F_in, &lin) )
        throw "file empty or starting with empty line";
    if     ( lin!=string("ASCII-96") )
        throw ".i96 file starts with illegal header [" + lin + "]";

    if ( !triv::freadln(F_in, &lin) || lin.length()<9 ||
         lin.substr(0,9)!=string("(a24,a56)") )
        throw "no valid intro to block 2";

    if ( !triv::freadln(F_in, &lin) || lin.length()<24 ||
         lin.substr(0,3)!=string("fil") )
        throw "tag 'fil' not found in block 2";
    if ( triv::strip(lin.substr(24))!=fshort ) {
                cout << "warning: i96 file '" << fshort <<
                    "' renamed from '" << lin.substr(24) << "'\n";
    }

    if ( !triv::freadln(F_in, &lin) || lin.length()<24 ||
         lin.substr(0,3)!=string("tit") )
        throw "tag 'tit' not found in block 2";
    fout->lDoc.push_back( triv::strip(lin.substr(24)) );

    while( !triv::freadln(F_in, &lin) && lin.length()>=24 &&
         lin.substr(0,3)==string("doc") )
        fout->lDoc.push_back( triv::strip(lin.substr(24)) );

    if( lin.substr(0,3)!=string("dir") )
        throw "tag 'dir' not found in block 2";

    while ( triv::freadln(F_in, &lin) &&
            !( lin.length()>=6 &&
               lin.substr(0,6)==string("&eob 2") ) )
        ;

    if ( !triv::freadln(F_in, &lin) )
        throw "missed &eob 3";

    if( lin.length()<9 || lin.substr(0,9)!=string("(a24,i16)") )
        throw "no valid intro to block 3";

    while ( !triv::freadln(F_in, &lin) &&
            !( lin.length()>=6 &&
               lin.substr(0,6)==string("&eob 3") ) )
        ;

    if ( !triv::freadln(F_in, &lin) )
        throw "missed &eob 3";

    if ( lin.length()<16 || lin.substr(0,16)!=string("(a24,a24,g20.10)") )
        throw "no valid intro to block 4";

    while ( !triv::freadln(F_in, &lin) && lin.length()>=48 &&
            lin.substr(0,4)!=string("&eob") ){
        if( sscanf( lin.substr(48).c_str(), "%lg", &r0 )!=1 )
            throw "no real value in rpar line";
        fout->RPar.push_back(
            CParam( lin.substr(0,24), lin.substr(24,24), r0 ) );
    }

    if( lin.substr(0,6)!=string("&eob 4") )
        throw "no valid eob 4";

    if ( !triv::freadln(F_in, &lin) || lin.length()<9 ||
         lin.substr(0,12)!=string("(a4,a24,a24)") )
        throw "no valid intro to block 5";

    if ( !triv::freadln(F_in, &lin) || lin.length()<28 ||
         lin.substr(0,1)!=string("x") )
        throw "no x coordinate";
    fout->xco = CCoord( triv::strip(lin.substr(4,24)),
                        triv::strip(lin.substr(28,24)) );

    if ( !triv::freadln(F_in, &lin) || lin.length()<28 ||
         lin.substr(0,1)!=string("y") )
        throw "no y coordinate";
    fout->yco = CCoord( triv::strip(lin.substr(4,24)),
                        triv::strip(lin.substr(28,24)) );

    if ( !triv::freadln(F_in, &lin) || lin.length()<28 ||
         lin.substr(0,1)!=string("z") )
        throw "no z coordinate";
    fout->ZCo.push_back( CCoord( triv::strip(lin.substr(4,24)),
                                 triv::strip(lin.substr(28,24)) ) );

    if ( !triv::freadln(F_in, &lin) || lin.length()<6 ||
         lin.substr(0,6)!=string("&eob 5") )
        throw "no valid eob 5";

    if ( !triv::freadln(F_in, &lin) || lin.length()<5 ||
         lin.substr(0,5)!=string("(a80)") )
        throw "no valid intro to block 6";

    while ( triv::freadln(F_in, &lin)>0 &&
            !( lin.length()>=6 &&
               lin.substr(0,6)==string("&eob 6") ) )
        fout->lDoc.push_back( lin );

    if ( !triv::freadln(F_in, &lin) )
        throw "missed &eob 6";
    for( int i=1; i<3; ++i )
        if ( !triv::freadln(F_in, &lin) )
            throw "no valid intro to block 7";

    if ( !triv::freadln(F_in, &lin) ||
         sscanf( lin.c_str(), "%i %i", &nspec, &ival )!=2 )
            throw "no nspec or 2nd arg in intro to block 7";

    for( int j=0; j<nspec; ++j ){
        sout = PSpec( new CSpec );
        if ( !triv::freadln(F_in, &lin) || lin.length()<10 ||
             lin.substr(0,9)!=string("&spectrum") ||
             sscanf( lin.substr(10).c_str(), "%i", &ival )!=1 ||
             ival!=j+1 )
            throw "no valid intro to spectrum";

        if ( !triv::freadln(F_in, &lin) ||
             sscanf( lin.c_str(), "%i %lg", &n, &r0 )!=2 )
            throw "no valid header in spectrum";
        sout->z.push_back( PObjDbl( new CObjDbl( r0 ) ) );

        for( int i=0; i<n; ++i ){
            if ( !F_in || !triv::freadln(F_in, &lin) )
                throw "i96: failed to read data line " +
                    S(i) + " of " + S(n);
            if ( sscanf( lin.c_str(), "%lg %lg %lg", &r0, &r1, &r2 )!=3 )
                throw "i96: bad data line: '" + lin + "'";
            sout->push_xy(r0, r1);
        }

        fout->V.push_back( move(sout) );
    }

    fout->as_on_disk = true;
    SMem::instance()->mem_store( move(fout) );
}
