//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  fsel.cpp
//! \brief Implement online file selection

#include "defs.hpp"

#include "olf.hpp"
#include "obj.hpp"
#include "mem.hpp"
#include "fsel.hpp"

//**************************************************************************************************
//*  Read online file selection
//**************************************************************************************************

//! Returns current selection.
const vector<int>& SFSel::sel() const
{
    return FSel;
}

//! Returns list of data files in current selection
vector<int> SFSel::selD() const
{
    vector<int> ret;
    for( int k: FSel )
        if( dynamic_cast<const COld*>(SMem::instance()->mem_get(k)) )
            ret.push_back( k );
    return ret;
}

//! Returns list of curve files in current selection
vector<int> SFSel::selC() const
{
    vector<int> ret;
    for( int k: FSel )
        if( dynamic_cast<const COlc*>(SMem::instance()->mem_get(k)) )
            ret.push_back( k );
    return ret;
}

//! First selected file.
COlo* SFSel::sel_first() const
{
    if( !FSel.size() )
        throw S("empty file selection");
    return SMem::instance()->mem_get(FSel[0]);
}

//! Maximum number of scans in selected files.
int SFSel::nJ_max() const
{
    int ret = 0;
    for( int iv: FSel )
        ret = std::max( ret, SMem::instance()->mem_get(iv)->nJ() );
    return ret;
}

//**************************************************************************************************
//*  Write to online file selection
//**************************************************************************************************

void SFSel::sel_push_back( int k )
{
    FSelNew.push_back( k );
}

//! Receive new file selection.
void SFSel::sel_push( const vector<int>& _FSel )
{
    FSelOld = FSel;
    FSel = _FSel;
}

//! Go back to previous file selection.
void SFSel::sel_delete()
{
    for (int i = FSel.size()-1; i>=0; --i)
        SMem::instance()->mem_del( FSel[i] );
    FSel = FSelOld;
    FSelOld.clear();
}

//! Return true if file selection can be taken from 'inp'.
void SFSel::set_selection( RObjVecInt _sel )
{
    sel_push( _sel->v );
}

//! Start collection of new files.
void SFSel::sel_collect_begin()
{   FSelNew.clear(); }

//! End collection of new files, and make them the new default selection.
void SFSel::sel_collect_end()
{
    if( !SMem::instance()->overwrite && FSelNew.size()>0 )
        sel_push( FSelNew );
}

//! Step up or down.
void SFSel::sel_increment( int step )
{
    for( int i=0; i<FSel.size(); ++i )
        FSel[i] += step;
}
