//****************************************************************************//
//* libtrivia: various little routines for C++                               *//
//* rng: random number generator wrapper                                     *//
//* (C) Joachim Wuttke, Antti Soininen 2001, 2012, 2013                      *//
//* http://apps.jcns.fz-juelich.de                                           *//
//****************************************************************************//

#include <gsl/gsl_randist.h>
#include <chrono>
#include <math.h>

#include "rng.hpp"

using namespace triv;

//****************************************************************************//
//  Implementation class
//****************************************************************************//

class triv::RandomNumberImplementation {
 public:
    RandomNumberImplementation() {
        r = gsl_rng_alloc( gsl_rng_default ); }
    ~RandomNumberImplementation() {
        gsl_rng_free( r ); }
    void reseed( unsigned long seed ) {
        gsl_rng_set( r, seed ); }
    double uniform() {
        return gsl_rng_uniform( r ); }
    double gaussian( double stdv ) {
        return gsl_ran_gaussian ( r, stdv ); }
    double cauchy( double width ) {
        return gsl_ran_cauchy ( r, width ); }
 private:
    gsl_rng *r;
};

//****************************************************************************//
//  Interface class
//****************************************************************************//

RandomNumberGenerator::RandomNumberGenerator()
{
    pimpl = new triv::RandomNumberImplementation();
}

RandomNumberGenerator::~RandomNumberGenerator()
{
    delete pimpl;
}

//! Set new random seed.

void RandomNumberGenerator::reseed( unsigned long seed )
{
    pimpl->reseed( seed );
}

//! Set new random seed from clock's nsec.

void RandomNumberGenerator::reseed_from_clock()
{
    std::chrono::nanoseconds time = std::chrono::system_clock::now().time_since_epoch();
    reseed( (unsigned long) time.count() );
}

double RandomNumberGenerator::uniform()
{
    return pimpl->uniform();
}

double RandomNumberGenerator::uniform( double rmin, double rmax )
{
    return rmin + (rmax-rmin)*uniform();
}

double RandomNumberGenerator::gaussian( double stdv, double cutoff )
{
    double ret;
    do {
        ret = pimpl->gaussian( stdv );
    } while( cutoff && fabs(ret)>cutoff );
    return ret;
}

double RandomNumberGenerator::cauchy( double width, double cutoff )
{
    double ret;
    do {
        ret = pimpl->cauchy( width );
    } while( cutoff && fabs(ret)>cutoff );
    return ret;
}
